from __future__ import division
import random

import pygame
from pygame.locals import *

from datacache import DataCache
from tile import Tile
import entity

class Level(object):
    
    def __init__(self, avatar):
        self.datacache = DataCache()
        self.image = pygame.Surface((800, 560), 0, pygame.display.get_surface())
        self.rect = self.image.get_rect()
        self.tiles = []
        self.doors = []
        tiles = self.generate_corridors([["floor" for i in range(20)] for j in range(14)], (0, 0, 19, 13))
        tiles = self.generate_doors(tiles)
        for i in xrange(len(tiles)):
            row = tiles[i]
            tile_row = []
            for j in xrange(len(row)):
                tile_name = row[j]
                tile = Tile(*self.datacache.get_data("tiles.py")[tile_name])
                tile.rect.topleft = (j*40, i*40)
                self.image.blit(tile.image, tile.rect)
                tile_row.append(tile)
            self.tiles.append(tile_row)
        
        self.entities = pygame.sprite.RenderUpdates()
        self.place_avatar(avatar)
        self.tesseract = entity.Tesseract(*self.datacache.get_data("entities.py")["tesseract"])
        self.place_tesseract(avatar)
        disturbances = self.datacache.get_data("disturbances.py")
        self.disturbances = []
        for n in range(random.randrange(1,4)):
            self.disturbances.append(disturbances[random.choice(disturbances.keys())])
    
    def update(self, interval):
        self.entities.update(interval)

    def full_draw(self):
        display = pygame.display.get_surface()
        display.blit(self.image, (0,0))
        self.entities.draw(display)
        pygame.display.update(Rect(0, 0, 800, 560))
    
    def draw(self):
        self.entities.clear(pygame.display.get_surface(), self.image)
        rects = self.entities.draw(pygame.display.get_surface())
        pygame.display.update(rects)
    
    def conditional_add_entity(self, entity, position, facing):
        entity.set_state(position, facing)
        if self.check_collision(entity):
            return False
        else:
            self.add_entity(entity)
            return True
        
    def add_entity(self, entity):
        self.entities.add(entity)
        for tile in self.get_entity_tiles(entity):
            tile.add_entity(entity)
        entity.level = self
                
    def remove_entity(self, entity):
        self.entities.remove(entity)
        for tile in self.get_entity_tiles(entity):
            tile.remove_entity(entity)
        entity.level = None
    
    def move_entity(self, entity, position, facing):
        old_position = entity.rect.center
        old_facing = entity.facing
        self.remove_entity(entity)
        entity.set_state(position, facing)
        collisions = self.check_collision(entity)
        if collisions:
            entity.set_state(old_position, old_facing)
            entity.collision(collisions)
        self.add_entity(entity)
                    
    def check_collision(self, entity):
        collisions = []
        if not self.rect.contains(entity.rect):
            collisions.append(self)
        tiles = self.get_entity_tiles(entity)
        for tile in tiles:
            collisions += tile.check_collision(entity)
        return collisions        
        
    def get_entity_tiles(self, entity):
        max_x = entity.rect.right // 40
        min_x = entity.rect.left // 40
        max_y = entity.rect.bottom // 40
        min_y = entity.rect.top // 40
        if max_y > 13: max_y = 13
        if max_x > 19: max_x = 19
        tiles = []
        for y in range(min_y, max_y + 1):
            for x in range(min_x, max_x + 1):
                tiles.append(self.tiles[y][x])
        return tiles
    
    def generate_corridors(self, tiles, area, direction = None):
        min_x, min_y, max_x, max_y = area
    
        if not direction:
            direction = random.choice(["vertical", "horizontal"])
    
        if direction == "vertical":
            if max_x - min_x < 7: 
                return tiles
            corridor_location = random.randrange(min_x + 3, max_x - 2)
            door1_location = random.randrange(min_y, max_y + 1)
            door2_location = random.randrange(min_y, max_y + 1)
            self.doors.append((corridor_location, door1_location))
            self.doors.append((corridor_location - 1, door1_location))
            self.doors.append((corridor_location - 2, door1_location))
            self.doors.append((corridor_location, door2_location))
            self.doors.append((corridor_location + 1, door2_location))
            self.doors.append((corridor_location + 2, door2_location))
            for row_index in range(min_y, max_y + 1):
                tiles[row_index][corridor_location - 1] = "wall"
                tiles[row_index][corridor_location + 1] = "wall"
            tiles = self.generate_corridors(tiles, (min_x, min_y, corridor_location - 2, max_y), "horizontal")
            return self.generate_corridors(tiles, (corridor_location + 2, min_y, max_x, max_y), "horizontal")
        else:
            if max_y - min_y < 7:
                return tiles
            corridor_location = random.randrange(min_y + 3, max_y - 2)
            door1_location = random.randrange(min_x, max_x + 1)
            door2_location = random.randrange(min_x, max_x + 1)
            self.doors.append((door1_location, corridor_location))
            self.doors.append((door1_location, corridor_location - 1))
            self.doors.append((door1_location, corridor_location - 2))
            self.doors.append((door2_location, corridor_location))
            self.doors.append((door2_location, corridor_location + 1))
            self.doors.append((door2_location, corridor_location + 2))
            for column_index in range(min_x, max_x + 1):
                tiles[corridor_location - 1][column_index] = "wall"
                tiles[corridor_location + 1][column_index] = "wall"
            tiles = self.generate_corridors(tiles, (min_x, min_y, max_x, corridor_location - 2), "vertical")
            return self.generate_corridors(tiles, (min_x, corridor_location + 2, max_x, max_y), "vertical")
    
    def generate_doors(self, tiles):
        for door in self.doors:
            door_x, door_y = door
            tiles[door_y][door_x] = "floor"
        return tiles
    
    def place_avatar(self, avatar):
        candidates = []
        for x in range(20):
            for y in range(14):
                if not self.tiles[y][x].impassable:
                    candidates.append((x*40 + 20,y*40 + 20))
        self.conditional_add_entity(avatar, random.choice(candidates), avatar.facing)
        
    def place_tesseract(self, avatar):
        candidates = []
        for x in range(20):
            for y in range(14):
                avatar_distance = ((avatar.position_x//40 - x)**2 + (avatar.position_y//40 - y)**2)**0.5
                if not self.tiles[y][x].impassable and avatar_distance > 10:
                    candidates.append((x*40 + 20,y*40 + 20))  
        self.conditional_add_entity(self.tesseract, random.choice(candidates), 0)                 